#include <iostream>
#include <string>
#include <sstream>

#include <cstdio>

#include <sys/stat.h>

#include <cups/raster.h>

#include "imagesource.h"
#include "imagesource_cms.h"
#include "imagesource_util.h"
#include "imagesource_solid.h"
#include "imagesource_montage.h"
#include "imagesource_dropshadow.h"
#include "imagesource_flatten.h"
#include "tiffsaver.h"
#include "jpegsaver.h"
#include "profilemanager.h"

using namespace std;

class ImageSource_CUPSRaster;

class CUPSRasterStream
{
	public:
	CUPSRasterStream() : ras(NULL)
	{
		ras=cupsRasterOpen(0,CUPS_RASTER_READ);
		if(!ras)
			throw "Can't open raster stream!";
	}
	~CUPSRasterStream()
	{
		if(ras)
			cupsRasterClose(ras);
	}
	protected:
	cups_raster_t *ras;
	friend class ImageSource_CUPSRaster;
};


class ImageSource_CUPSRaster : public ImageSource
{
	public:
	ImageSource_CUPSRaster(CUPSRasterStream &stream) : ImageSource(), stream(stream), buffer(NULL)
	{
		if(cupsRasterReadHeader2(stream.ras,&header))
		{
			width=header.cupsWidth;
			height=header.cupsHeight;
			xres=header.HWResolution[0];
			yres=header.HWResolution[1];
			randomaccess=false;
			std::cerr << "Got colour space: " << header.cupsColorSpace << endl;
			switch(header.cupsColorSpace)
			{
				case CUPS_CSPACE_RGB:
					type=IS_TYPE_RGB;
					samplesperpixel=3;
					break;
				case CUPS_CSPACE_RGBA:
					type=IS_TYPE_RGBA;
					samplesperpixel=4;
					break;
				case CUPS_CSPACE_CMYK:
					type=IS_TYPE_CMYK;
					samplesperpixel=4;
					break;
				default:
					throw "Image type currently unsupported";
					break;
			}

			/* allocate memory for 1 line */
			buffer = new unsigned char[header.cupsBytesPerLine+1];

			MakeRowBuffer();
			currentrow=-1;
		}
		else
			throw 0;
	}
	~ImageSource_CUPSRaster()
	{
		for(int extra=currentrow;extra<height;++extra)
		{
			GetRow(extra);
		}
		if(buffer)
			delete[] buffer;
	}
	ISDataType *GetRow(int row)
	{
		if(currentrow==row)
			return(rowbuffer);

		int read=0;
		while(currentrow<row)
		{
			read=cupsRasterReadPixels(stream.ras, buffer, header.cupsBytesPerLine);
			++currentrow;
		}
		if(read==0)
			throw "Error while reading image";

		ISDataType *dst=rowbuffer;
		unsigned char *src=buffer;

		switch(header.cupsBitsPerPixel)
		{
			case 24:
			case 32:
				for(int x=0;x<width*samplesperpixel;++x)
				{
					int t=*src++;
					*dst++=EIGHTTOIS(t);
				}
				break;
			case 48:
			case 64:
				for(int x=0;x<width*samplesperpixel;++x)
				{
					int t=*src++;
					*dst++=(t<<8)|*src++;
				}
				break;
		}
		currentrow=row;
		return(rowbuffer);
	}
	protected:
	cups_page_header2_t header;
	CUPSRasterStream &stream;
	unsigned char *buffer;
};


ISDeviceNValue ParseWebColour(const char *col)
{
	if(!col || strlen(col)<7 || col[0]!='#')
		throw "Invalid colour specification";
	int t,r,g,b;
	t=col[1]-'0'; if(t>9) t=(col[1]|32)-'a'+10;
	r=t*16;
	t=col[2]-'0'; if(t>9) t=(col[2]|32)-'a'+10;
	r+=t;

	t=col[3]-'0'; if(t>9) t=(col[3]|32)-'a'+10;
	g=t*16;
	t=col[4]-'0'; if(t>9) t=(col[4]|32)-'a'+10;
	g+=t;

	t=col[5]-'0'; if(t>9) t=(col[5]|32)-'a'+10;
	b=t*16;
	t=col[6]-'0'; if(t>9) t=(col[6]|32)-'a'+10;
	b+=t;

	ISDeviceNValue result(3,0);
	result[0]=EIGHTTOIS(r);
	result[1]=EIGHTTOIS(g);
	result[2]=EIGHTTOIS(b);
	return(result);
}


class Settings : public ConfigFile, public ConfigDB
{
	public:
	Settings() : ConfigFile(), ConfigDB(Template)
	{
		new ConfigDBHandler(this,"[RasterToTIFF]",this);
	}
	~Settings()
	{
	}
	protected:
	static ConfigTemplate Template[];
};


ConfigTemplate Settings::Template[]=
{
	ConfigTemplate("OutPath","/var/www/spool"),
	ConfigTemplate("BGColour","#f4f4f4"),
	ConfigTemplate()
};


int main(int argc,char **argv)
{
	CUPSRasterStream stream;
	try
	{
		Settings conf;
		ProfileManager pm(&conf,"[Colour Management]");
		conf.ParseConfigFile("/etc/rastertotiff.conf");
		const char *outpath=conf.FindString("OutPath");
		ISDeviceNValue bgcolour=ParseWebColour(conf.FindString("BGColour"));

		while(1)
		{
			// construct suitable output filenames
			time_t seconds=time(NULL);
			std::stringstream filenamestream;
			filenamestream << outpath << "/" << seconds;

			std::string filename = filenamestream.str() + ".tif";	// Image...
			std::string thumbnail_fn = filenamestream.str() + "_tn.jpg";	// ... and thumbnail

			{
				ImageSource *is=new ImageSource_CUPSRaster(stream);
				TIFFSaver tiffsaver(filename.c_str(),ImageSource_rp(is));
				tiffsaver.Save();
			}
			chmod(filename.c_str(),S_IREAD | S_IWRITE | S_IRGRP | S_IROTH);

			// Image is saved, now construct a thumbnail...

			ImageSource *is=ISLoadImage(filename.c_str());
			int w=256;
			int h=(is->height*256)/is->width;
			if(h>256)
			{
				w=(is->width*256)/is->height;
				h=256;
			}
			is=ISScaleImageBySize(is,w,h,IS_SCALING_DOWNSAMPLE);

			if(is->type==IS_TYPE_CMYK)
			{
				// Convert the thumbnail to RGB...
				RefCountPtr<CMSProfile> cmyk(pm.GetDefaultProfile(IS_TYPE_CMYK));
				RefCountPtr<CMSProfile> rgb(pm.GetDefaultProfile(IS_TYPE_CMYK));
				RefCountPtr<CMSTransform> trans(new CMSTransform(&*rgb,&*cmyk,LCMSWRAPPER_INTENT_RELATIVE_COLORIMETRIC));
				is=new ImageSource_CMS(is,trans);
			}
			is=new ImageSource_DropShadow(is,8,4);

			ImageSource *background=new ImageSource_Solid(IS_TYPE_RGB,is->width,is->height,&bgcolour[0]);
			ImageSource_Montage *mon2=new ImageSource_Montage(IS_TYPE_RGB);
			mon2->Add(is,0,0);
			mon2->Add(background,0,0);
			is=new ImageSource_Flatten(mon2);
			JPEGSaver jpegsaver(thumbnail_fn.c_str(),ImageSource_rp(is));
			jpegsaver.Save();
			chmod(thumbnail_fn.c_str(),((S_IREAD|S_IWRITE)) | ((S_IREAD|S_IWRITE)>>3) | ((S_IREAD|S_IWRITE) >>6));
		}
	}
	catch(int)
	{
		// No more images in stream
	}
	catch(const char *err)
	{
		std::cerr << err << std::endl;
	}

	return(0);
}

